from test.test_support import run_unittest
import sys
import xdis.std as dis
import StringIO

# Minimal tests for dis module

import unittest

def _f(a):
    print a
    return 1

dis_f = """\
%3d:           0 LOAD_FAST                 0 (a)
               3 PRINT_ITEM
               4 PRINT_NEWLINE

%3d:           5 LOAD_CONST                1 (1)
               8 RETURN_VALUE

"""%(_f.func_code.co_firstlineno + 1,
     _f.func_code.co_firstlineno + 2)


def bug708901():
    for res in range(1,
                     10):
        pass

dis_bug708901 = """\
%3d:           0 SETUP_LOOP               23 (to 26)
               3 LOAD_GLOBAL               0 (range)
               6 LOAD_CONST                1 (1)

%3d:           9 LOAD_CONST                2 (10)
              12 CALL_FUNCTION             2 (2 positional, 0 keyword pair)
              15 GET_ITER
         >>   16 FOR_ITER                  6 (to 25)
              19 STORE_FAST                0 (res)

%3d:          22 JUMP_ABSOLUTE            16 (to 16)
         >>   25 POP_BLOCK
         >>   26 LOAD_CONST                0 (None)
              29 RETURN_VALUE

"""%(bug708901.func_code.co_firstlineno + 1,
     bug708901.func_code.co_firstlineno + 2,
     bug708901.func_code.co_firstlineno + 3)

class DisTests(unittest.TestCase):
    def do_disassembly24(self, func, expected):
        s = StringIO.StringIO()
        save_stdout = sys.stdout
        sys.stdout = s
        dis.dis(func)
        sys.stdout = save_stdout
        got = s.getvalue()
        # Trim trailing blanks (if any).
        lines = got.split('\n')
        lines = [line.rstrip() for line in lines]
        expected = expected.split("\n")
        import difflib
        if expected != lines:
            self.fail(
                "events did not match expectation:\n" +
                "\n".join(difflib.ndiff(expected,
                                        lines)))

    def test_opmap(self):
        self.assertEqual(dis.opmap["STOP_CODE"], 0)
        self.assertEqual(dis.opmap["LOAD_CONST"] in dis.hasconst, True)
        self.assertEqual(dis.opmap["STORE_NAME"] in dis.hasname, True)

    def test_opname(self):
        self.assertEqual(dis.opname[dis.opmap["LOAD_FAST"]], "LOAD_FAST")

    def test_boundaries(self):
        self.assertEqual(dis.opmap["EXTENDED_ARG"], dis.EXTENDED_ARG)
        self.assertEqual(dis.opmap["STORE_NAME"], dis.HAVE_ARGUMENT)

    def test_dis(self):
        self.do_disassembly24(_f, dis_f)

    def test_bug_708901(self):
        self.do_disassembly24(bug708901, dis_bug708901)

def test_main():
    run_unittest(DisTests)


if __name__ == "__main__":
    test_main()



# Copyright (c) 2016-2017 by Rocky Bernstein
"""
CPython independent disassembly routines

There are two reasons we can't use Python's built-in routines
from dis. First, the bytecode we are extracting may be from a different
version of Python (different magic number) than the version of Python
that is doing the extraction.

Second, we need structured instruction information for the
(de)-parsing step. Python 3.4 and up provides this, but we still do
want to run on Python 2.7.
"""

# Note: we tend to eschew new Python 3 things, and even future
# imports so this can run on older Pythons. This is
# intended to be a more cross-version Python program

import datetime, sys

try:
    from collections import deque
except ImportError:
    class deque:
        def __init__(self, initial_todo=[]):
            self.todo = initial_todo

        def popleft(self):
            return self.todo.pop()

        def append(self, item):
            return self.todo.append(item)

        def __len__(self):
            return len(self.todo)

import xdis

from xdis import IS_PYPY
from xdis.bytecode import Bytecode
from xdis.code import iscode
from xdis.load import check_object_path, load_module
from xdis.util import format_code_info
from xdis.version import VERSION
from xdis.op_imports import op_imports

def get_opcode(version, is_pypy):
    # Set up disassembler with the right opcodes
    lookup = str(version)
    if is_pypy:
        lookup += 'pypy'
    if lookup in op_imports.keys():
        return op_imports[lookup]
    if is_pypy:
        pypy_str = ' for pypy'
    else:
        pypy_str = ''
    raise TypeError("%s is not a Python version%s I know about" %
                    (version, pypy_str))

def disco(bytecode_version, co, timestamp, out=sys.stdout,
          is_pypy=False, magic_int=None, source_size=None,
          header=True, asm_format=False, dup_lines=False):
    """
    diassembles and deparses a given code block 'co'
    """

    assert iscode(co)

    # store final output stream for case of error
    real_out = out or sys.stdout
    if is_pypy:
        co_pypy_str = 'PyPy '
    else:
        co_pypy_str = ''
    if IS_PYPY:
        run_pypy_str = 'PyPy '
    else:
        run_pypy_str = ''
    if magic_int:
        m = str(magic_int)
    else:
        m = ""
    real_out.write(('# pydisasm version %s\n# %sPython bytecode %s%s'
               '\n# Disassembled from %sPython %s\n') %
              (VERSION, co_pypy_str, bytecode_version,
               " (%s)" % m, run_pypy_str,
               '\n# '.join(sys.version.split('\n'))))
    if timestamp > 0:
        value = datetime.datetime.fromtimestamp(timestamp)
        real_out.write('# Timestamp in code: %d' % timestamp)
        real_out.write(value.strftime(' (%Y-%m-%d %H:%M:%S)\n')
)
    if source_size:
        real_out.write('# Source code size mod 2**32: %d bytes\n' % source_size)

    if co.co_filename and not asm_format:
        real_out.write(format_code_info(co, bytecode_version) + "\n")
        pass

    opc = get_opcode(bytecode_version, is_pypy)

    if asm_format:
        disco_loop_asm_format(opc, bytecode_version, co, real_out)
    else:
        queue = deque([co])
        disco_loop(opc, bytecode_version, queue, real_out)


def disco_loop(opc, version, queue, real_out, dup_lines=False):
    """Disassembles a queue of code objects. If we discover
    another code object which will be found in co_consts, we add
    the new code to the list. Note that the order of code discovery
    is in the order of first encountered which is not amenable for
    the format used by a disassembler where code objects should
    be defined before using them in other functions.
    However this is not recursive and will overall lead to less
    memory consumption at run time.
    """

    while len(queue) > 0:
        co = queue.popleft()
        if co.co_name != '<module>':
            real_out.write("\n" + format_code_info(co, version) + "\n")

        bytecode = Bytecode(co, opc, dup_lines=dup_lines)
        real_out.write(bytecode.dis() + "\n")

        for c in co.co_consts:
            if iscode(c):
                queue.append(c)
            pass
        pass

def disco_loop_asm_format(opc, version, co, real_out):
    """Produces disassembly in a format more conducive to
    automatic assembly by producing inner modules before they are
    used by outer ones. Since this is recusive, we'll
    use more stack space at runtime.
    """
    for c in co.co_consts:
        if iscode(c):
            disco_loop_asm_format(opc, version, c, real_out)
        pass

    if co.co_name != '<module>' or co.co_filename:
        real_out.write("\n" + format_code_info(co, version) + "\n")

    bytecode = Bytecode(co, opc, dup_lines=True)
    real_out.write(bytecode.dis(asm_format=True) + "\n")

def disassemble_file(filename, outstream=sys.stdout, asm_format=False):
    """
    disassemble Python byte-code file (.pyc)

    If given a Python source file (".py") file, we'll
    try to find the corresponding compiled object.
    """
    filename = check_object_path(filename)
    version, timestamp, magic_int, co, is_pypy, source_size  = load_module(filename)
    disco(version, co, timestamp, outstream, is_pypy, magic_int, source_size,
          asm_format=asm_format)
    # print co.co_filename
    return filename, co, version, timestamp, magic_int

def _test():
    """Simple test program to disassemble a file."""
    argc = len(sys.argv)
    if argc != 2:
        if argc == 1 and xdis.PYTHON3:
            fn = __file__
        else:
            sys.stderr.write("usage: %s [-|CPython compiled file]\n" % __file__)
            sys.exit(2)
    else:
        fn = sys.argv[1]
    disassemble_file(fn)

if __name__ == "__main__":
    _test()



import unittest
from xdis import IS_PYPY, PYTHON_VERSION
from xdis.main import get_opcode
import dis

class TestOpcode(unittest.TestCase):

    def test_opcode(self):
        opc = get_opcode(PYTHON_VERSION, IS_PYPY)
        opmap = dict([(k.replace('+', '_'), v)
                      for (k, v) in dis.opmap.items()])
        # print(set(opmap.items()) - set(l['opmap'].items()))
        # print(set(l['opmap'].items()) - set(opmap.items()))

        for item in opc.opmap.items():
            assert item in opmap.items()
            for item in opmap.items():
                self.assertTrue(item in opc.opmap.items())

        fields = """hascompare hasconst hasfree hasjabs hasjrel haslocal
        hasname""".split()
        for field in fields:
            opc_set = set(getattr(opc, field))
            dis_set = set(getattr(dis, field))
            self.assertTrue(opc_set == dis_set,
                            ("diff in %s: %s" %
                             (field, ', '.join([opc.opname[i]
                                                for i in list(opc_set ^ dis_set)]))))
            pass
        return

if __name__ == '__main__':
    unittest.main()



from xdis.bytecode import offset2line

import unittest

class Offset2LineTests(unittest.TestCase):

    def test_offset2line(self):
        ary = ((20, 1), (40, 10), (60, 45))
        for offset, expect in (
                (61, 45),
                (20, 1),
                (40, 10),
                (39, 1),
                (41, 10),
                (60, 45),
                (21, 1)):
            got = offset2line(offset, ary)
            self.assertEqual(got, expect,
                             ("offset=%d, got=%d, expect=%d" % (offset, got, expect)))
        ary = ((20, 1), (50, 10))
        for offset, expect in (
                (10, 0),
                (55, 10),
                (20, 1),
                (50, 10)):
            got = offset2line(offset, ary)
            self.assertEqual(got, expect,
                             ("offset=%d, got=%d, expect=%d" % (offset, got, expect)))
        ary = ((25, 1),)
        for offset, expect in (
                (10, 0),
                (30, 1),
                (25, 1)):
            got = offset2line(offset, ary)
            self.assertEqual(got, expect,
                             ("offset=%d, got=%d, expect=%d" % (offset, got, expect)))
        ary = tuple()
        for offset, expect in ((10, 0),):
            got = offset2line(offset, ary)
            self.assertEqual(got, expect,
                             ("offset=%d, got=%d, expect=%d" % (offset, got, expect)))
            pass
        return

if __name__ == "__main__":
    unittest.main()
